// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.*;
import java.util.concurrent.CyclicBarrier;

import common.Tupel;
import common.Vec3;
import model.*;

public class BCCTestStructure extends BCCStructure {

	private static final int SF = 8;
	private static final int BCC = 0;
	private static final int FCC = 1;
	
	public BCCTestStructure() {
		super();
		this.minRBVLength.defaultValue = 0.1f;
	}
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new BCCTestStructure();
	}
	
	@Override
	protected String getIDName() {
		return "BCC_Test";
	}
	
	@Override
	public int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb) {
		ArrayList<Vec3> neigh = nnb.getNeighVec(atom);
		/*
		 * type=0: bcc
		 * type=1: fcc
		 * type=2: hcp
		 * type=3: 14 neighbors
		 * type=4: 11-13 neighbors
		 * type=5: 15 neighbors
		 * type=6: less than 11 neighbors
		 * type=7: unknown
		 */
		if (neigh.size() < 11) return 6;
		else if (neigh.size() == 11) return 4;
		else if (neigh.size() == 13) {
			int co_x0 = 0;
			int co_x1 = 0;
			int co_x2 = 0;
			
			for (int i = 0; i < neigh.size(); i++) {
				Vec3 v = neigh.get(i);
				float v_length = v.getLength();
				
				for (int j = 0; j < i; j++) {
					Vec3 u = neigh.get(j);
					float u_length = u.getLength();
					float a = v.dot(u) / (v_length*u_length);
					
					if (a < -.975)
						co_x0++;
					if (a < -0.93 &&  a > -0.975)
						co_x1++;
					
					if (a < 0.57 &&  a > 0.425)
						co_x2++;
				}
			}
			if (co_x0==3 && co_x1 == 3 && co_x2 >= 3)
				return SF; 
			
			return 4;
		}
		else if (neigh.size() == 15) return 5;
		else if (neigh.size() > 15) return 7;
		else {
			int co_x0 = 0;
			int co_x1 = 0;
			int co_x2 = 0;
			for (int i = 0; i < neigh.size(); i++) {
				Vec3 v = neigh.get(i);
				float v_length = v.getLength();
				
				for (int j = 0; j < i; j++) {
					Vec3 u = neigh.get(j);
					float u_length = u.getLength();
					float a = v.dot(u) / (v_length*u_length);
					
					if (a < -.945)
						co_x0++;
					else if (a < -.915)
						co_x1++;
					else if (a > -.75 && a< -0.67)
						co_x2++;
				}
			}
			
			if (co_x0 > 5 && co_x0+co_x1==7 && co_x2<=2 && neigh.size()==14) return BCC;
			else if (co_x0 == 6 && neigh.size() == 12) return FCC;
			else if (co_x0 == 3 && neigh.size() == 12) return 2;
			else if (neigh.size() == 12) return 4;
			else return 3;
		}
	}
	
	
	@Override
	public void identifyDefectAtoms(List<Atom> atoms, NearestNeighborBuilder<Atom> nnb, 
			int start, int end, CyclicBarrier barrier) {
		for (int i=start; i<end; i++){
			if (Thread.interrupted()) return;
			
			if ((i-start)%10000 == 0)
				Configuration.currentFileLoader.getProgressMonitor().addToCounter(10000);
			
			Atom a = atoms.get(i);
			int type = identifyAtomType(a, nnb); 
			//Relabel
			a.setType(type);
		}
		
		Configuration.currentFileLoader.getProgressMonitor().addToCounter((end-start)%10000);
		try {
			barrier.await();
		} catch (Exception e) {
			if (Thread.interrupted()) return;
		}
	
		ArrayList<Atom> sfRelabel = new ArrayList<Atom>();
		
		//Test if atoms assigned as stacking faults are really stacking faults
		for (int i=start; i<end; i++){
			if (Thread.interrupted()) return;
			Atom a = atoms.get(i);
			
			if (a.getType() == SF){
				int countSFPairs = 0;
				
				ArrayList<Tupel<Atom, Vec3>> neiAndVec = nnb.getNeighAndNeighVec(a);
				for (int j=0; j<neiAndVec.size()-1; j++){
					Atom b = neiAndVec.get(j).getO1();
					if (b.getType() == SF){
						for (int k=j; k<neiAndVec.size(); k++){
							Atom c = neiAndVec.get(k).getO1();
							if (c.getType() == SF && 
									b.subClone(a).getAngle(c.subClone(a))>0)
								countSFPairs++;
						}
					}
				}
				
				if (countSFPairs <= 8) sfRelabel.add(a);
			}
		}
		
		for (Atom a : sfRelabel)
			a.setType(4);
	}
	
	@Override
	public List<Atom> getStackingFaultAtoms(AtomData data){
		ArrayList<Atom> sfAtoms = new ArrayList<Atom>();
		for (int i=0; i<data.getAtoms().size(); i++){
			Atom a = data.getAtoms().get(i);
			if (a.getType() == SF && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN))
				sfAtoms.add(a);
		}
		return sfAtoms;
	}
	
	@Override
	public boolean hasStackingFaults() {
		return true;
	}
	
	
	@Override
	public Vec3[] getStackingFaultNormals(CrystalRotationTools crt){
		Vec3[] stackingFaultNormals = new Vec3[6];
		
		stackingFaultNormals[0] = crt.getInXYZ(new Vec3( 1f, 0f, 1f)).normalize(); 
		stackingFaultNormals[1] = crt.getInXYZ(new Vec3(-1f, -1f,0f)).normalize();
		stackingFaultNormals[2] = crt.getInXYZ(new Vec3( 1f,-1f, 0f)).normalize();
		stackingFaultNormals[3] = crt.getInXYZ(new Vec3( 0f,-1f, 1f)).normalize();
		stackingFaultNormals[4] = crt.getInXYZ(new Vec3(-1f, 0f, 1f)).normalize();
		stackingFaultNormals[5] = crt.getInXYZ(new Vec3( 0f, 1f, 1f)).normalize();
		
		return stackingFaultNormals;
	}
	
	@Override
	public int getSurfaceType() {
		return 6;
	}
	
	@Override
	public boolean isRBVToBeCalculated(Atom a) {
		if ( (ImportStates.POLY_MATERIAL.isActive() && a.getGrain() == Atom.IGNORED_GRAIN)) return false;
		int type = a.getType();		
		if (type != 0 && type <= 6) {			
			return true;
		}
		return false;		
	}
	
	@Override
	public int getNumberOfTypes() {
		return 9;
	}
	
	@Override
	public String getNameForType(int i) {
		switch (i) {
			case 0: return "bcc";
			case 1: return "fcc";
			case 2: return "hcp";
			case 3: return "14 neighbors";
			case 4: return "11-13 neighbors";
			case 5: return "15 neighbors";
			case 6: return "<11 neighbors";
			case 7: return ">15 neighbors";
			case 8: return "stacking fault";
			default: return "unknown";
		}
	}
}
